#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause OR Apache-2.0
# encoding: utf-8

import argparse
import logging

import cbor2

from ud3tn_utils.aap2 import (
    AAP2TCPClient,
    AAP2UnixClient,
    AuthType,
    BundleADU,
    BundleADUFlags,
    ResponseStatus,
)
from ud3tn_utils.aap2.bin.helpers import (
    add_common_parser_arguments,
    get_secret_from_args,
    initialize_logger,
)
from pyd3tn.bundle7 import (
    BibeProtocolDataUnit,
    Bundle,
    PayloadBlock,
    PrimaryBlock,
)


logger = logging.getLogger(__name__)


def build_bpdu(inner_source, inner_dest, payload):
    """Encapsulates a regular bundle with the chosen payload
    in a BIBE Administrative Record, thus forming a BIBE Bundle.

    Args:
        inner_source (str): Source EID of the encapsulated bundle
        inner_dest (str): Destination EID of the encapsulated bundle
        payload (bytes): The payload of the encapsulated bundle

    Returns:
        BPDU (bytes): The bytes making up the BPDU
    """
    inner_bundle = Bundle(
        PrimaryBlock(
            destination=inner_dest,
            source=inner_source,
        ),
        PayloadBlock(payload)
    )
    bibe_ar = BibeProtocolDataUnit(
        bundle=inner_bundle,
        transmission_id=0,
        retransmission_time=0,
        compatibility=False,
    )

    return cbor2.dumps(bibe_ar.record_data)


def main():
    import sys

    parser = argparse.ArgumentParser(
        description="send a bundle via uD3TN's AAP interface",
    )
    add_common_parser_arguments(parser)
    parser.add_argument(
        "dest_eid",
        help="the destination EID of the created bundle",
    )
    parser.add_argument(
        "PAYLOAD",
        default=None,
        nargs="?",
        help="the payload of the created bundle, (default: read from STDIN)",
    )
    parser.add_argument(
        "--bdm-auth",
        action="store_true",
        help=("request the BUNDLE_ADU_WITH_BDM_AUTH flag to be set, e.g., to "
              "manage BDMs (needs the administrative secret to be set)"))
    parser.add_argument(
        "--bibe-source",
        default=None,
        help=("if set, the payload will be encapsulated twice using BIBE, "
              "with the source of the inner bundle set to the given EID"))
    parser.add_argument(
        "--bibe-destination",
        default=None,
        help=("if set, the payload will be encapsulated twice using BIBE, "
              "with the inner bundle addressed to the given EID"))
    args = parser.parse_args()

    if args.PAYLOAD:
        payload = args.PAYLOAD.encode("utf-8")
    else:
        payload = sys.stdin.buffer.read()
        sys.stdin.buffer.close()

    global logger
    logger = initialize_logger(args.verbosity)

    is_bibe = (
        args.bibe_source is not None or
        args.bibe_destination is not None
    )
    if is_bibe and (args.bibe_source is None or args.bibe_destination is None):
        logger.fatal("--bibe-source and --bibe-destination must both be set")
        sys.exit(1)

    if args.tcp:
        aap2_client = AAP2TCPClient(address=args.tcp)
    else:
        aap2_client = AAP2UnixClient(address=args.socket)
    with aap2_client:
        secret = aap2_client.configure(
            args.agentid,
            subscribe=False,
            secret=get_secret_from_args(args),
            auth_type=(
                AuthType.AUTH_TYPE_DEFAULT if not args.bdm_auth
                else AuthType.AUTH_TYPE_BUNDLE_DISPATCH
            ),
        )
        logger.info("Assigned agent secret: '%s'", secret)
        if is_bibe:
            payload = build_bpdu(
                args.bibe_source,
                args.bibe_destination,
                payload,
            )
            flags = [BundleADUFlags.BUNDLE_ADU_BPDU]
        else:
            flags = [BundleADUFlags.BUNDLE_ADU_NORMAL]
        if args.bdm_auth:
            flags += [BundleADUFlags.BUNDLE_ADU_WITH_BDM_AUTH]
        aap2_client.send_adu(
            BundleADU(
                dst_eid=args.dest_eid,
                payload_length=len(payload),
                adu_flags=flags,
            ),
            payload,
        )
        assert (
            aap2_client.receive_response().response_status ==
            ResponseStatus.RESPONSE_STATUS_SUCCESS
        )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: AGPL-3.0-or-later
# encoding: utf-8

"""Minimal Bundle Dispatcher service with static routing for uD3TN-NG."""

import argparse
import json
import logging
import sys

from pyd3tn.eid import get_node_id

from ud3tn_utils.aap2 import (
    AAP2UnixClient,
    AAP2TCPClient,
    AAP2ServerDisconnected,
    AAPResponse,
    AuthType,
    DispatchReason,
    DispatchResult,
    ResponseStatus,
)
from ud3tn_utils.aap2.bin.helpers import (
    add_common_parser_arguments,
    add_keepalive_parser_argument,
    get_secret_from_args,
    initialize_logger,
)

# Which DispatchReason values we accept for bundles to be forwarded
# (in other cases they are dropped).
VALID_DISPATCH_REASONS = (
    DispatchReason.DISPATCH_REASON_NO_FIB_ENTRY,
)


logger = logging.getLogger(__name__)


def run_static_bdm(aap2_client, routing_dict):
    logger.info("Waiting for dispatch event...")

    while True:
        try:
            msg = aap2_client.receive_msg()
        except AAP2ServerDisconnected:
            logger.warning("µD3TN has closed the connection.")
            sys.exit(1)
        except KeyboardInterrupt:
            logger.info("Terminated by keyboard interrupt.")
            sys.exit(130)  # exit status for SIGINT

        if msg.WhichOneof("msg") == "keepalive":
            aap2_client.send_response_status(
                ResponseStatus.RESPONSE_STATUS_ACK,
            )
            continue
        elif msg.WhichOneof("msg") != "dispatch_event":
            logger.info("Received message with field '%s' set, discarding.",
                        msg.WhichOneof("msg"))
            aap2_client.send_response_status(
                ResponseStatus.RESPONSE_STATUS_SUCCESS,
            )
            continue

        if msg.dispatch_event.reason not in VALID_DISPATCH_REASONS:
            logger.info("Received dispatch event with reason %d, dropping.",
                        msg.dispatch_event.reason)
            aap2_client.send_response_status(
                ResponseStatus.RESPONSE_STATUS_SUCCESS,
            )
            continue

        # Main dispatch logic
        bdl = msg.dispatch_event.bundle
        dst_node_id = get_node_id(bdl.dst_eid)
        next_hop_node_id = routing_dict.get(
            dst_node_id,
            # default: assume identity (destination must be next hop)
            dst_node_id,
        )
        logger.info("Determined next hop '%s' for bundle with dst = '%s'.",
                    next_hop_node_id, bdl.dst_eid)
        disp_result = DispatchResult(
            next_hops=[DispatchResult.NextHopEntry(node_id=next_hop_node_id)]
        )
        aap2_client.send(AAPResponse(
            response_status=ResponseStatus.RESPONSE_STATUS_SUCCESS,
            dispatch_result=disp_result,
        ))


def main():
    parser = argparse.ArgumentParser(
        description="bundle dispatcher module using a static table",
    )
    add_common_parser_arguments(parser)
    add_keepalive_parser_argument(parser)
    parser.add_argument(
        "routing_table_file",
        help="the JSON file to determine the next hop for a given bundle",
    )

    args = parser.parse_args()
    global logger
    logger = initialize_logger(args.verbosity)

    with open(args.routing_table_file) as f:
        routing_dict = json.load(f)

    if args.tcp:
        aap2_client = AAP2TCPClient(address=args.tcp)
    else:
        aap2_client = AAP2UnixClient(address=args.socket)
    with aap2_client:
        secret = aap2_client.configure(
            args.agentid,
            subscribe=True,
            secret=get_secret_from_args(args),
            auth_type=AuthType.AUTH_TYPE_BUNDLE_DISPATCH,
            keepalive_seconds=args.keepalive_seconds,
        )
        logger.info("AAP 2.0 agent registered and configured as BDM!")
        logger.info("Assigned agent secret: '%s'", secret)
        run_static_bdm(aap2_client, routing_dict)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause OR Apache-2.0
# encoding: utf-8

"""
A tool that sends received bundles back to the sender.
"""

import argparse
import logging
import sys

from ud3tn_utils.aap import AAPUnixClient, AAPTCPClient, AAPMessage
from ud3tn_utils.aap.aap_message import AAPMessageType
from ud3tn_utils.aap.bin.helpers import (
    add_common_parser_arguments,
    initialize_logger,
)


logger = logging.getLogger(__name__)


def run_echo(aap_client):

    logger.info("Registered EID '%s', waiting for bundles...", aap_client.eid)

    while True:

        try:
            msg = aap_client.receive()
        except KeyboardInterrupt:
            logger.info("Terminated by keyboard interrupt.")
            sys.exit(130)  # exit status for SIGINT
        if not msg:
            logger.warning("µD3TN has closed the connection.")
            sys.exit(1)

        if msg.msg_type == AAPMessageType.RECVBUNDLE:
            logger.debug(
                "Received bundle of length %d bytes from '%s'",
                len(msg.payload),
                msg.eid,
            )
            aap_client.socket.send(
                AAPMessage(
                    AAPMessageType.SENDBUNDLE,
                    msg.eid,
                    msg.payload
                ).serialize()
            )
            logger.debug("Sent a bundle back to '%s'", msg.eid)


def main():
    parser = argparse.ArgumentParser(
        description="sends received bundles back to the sender",
    )

    add_common_parser_arguments(parser)

    args = parser.parse_args()
    global logger
    logger = initialize_logger(args.verbosity + 1)  # log INFO by default

    if args.tcp:
        addr = (args.tcp[0], int(args.tcp[1]))
        with AAPTCPClient(address=addr) as aap_client:
            aap_client.register(args.agentid)
            run_echo(aap_client)
    else:
        with AAPUnixClient(address=args.socket) as aap_client:
            aap_client.register(args.agentid)
            run_echo(aap_client)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: AGPL-3.0-or-later
# encoding: utf-8

"""Minimal Bundle Dispatcher service with static routing for uD3TN-NG."""

import argparse
import asyncio
import dataclasses
import logging
import re
import sys
import time

from typing import Dict, List, Optional, Set, Tuple

from google.protobuf.internal import encoder as proto_encoder

from pyd3tn.eid import get_node_id

from ud3tn_utils.aap2 import (
    AAP2AsyncClient,
    AAP2AsyncUnixClient,
    AAP2AsyncTCPClient,
    AAP2ServerDisconnected,
    AAPMessage,
    AAPResponse,
    AuthType,
    Bundle,
    BundleADU,
    BundleADUFlags,
    BundleDispatchInfo,
    DispatchReason,
    DispatchEvent,
    DispatchResult,
    Link,
    LinkStatus,
    ResponseStatus,
)
from ud3tn_utils.config import (
    UNIX_TO_DTN_OFFSET,
)
from ud3tn_utils.storage_agent import (
    CompoundBundleId,
    StorageCall,
    StorageOperation,
)
from ud3tn_utils.aap2.bin.helpers import (
    add_common_parser_arguments,
    add_keepalive_parser_argument,
    get_secret_from_args,
    initialize_logger,
    DEFAULT_CONFIG_AGENT_ID_DTN,
    DEFAULT_CONFIG_AGENT_ID_IPN,
)

CONFIG_PARSE_REGEX = re.compile(
    r"^(?P<cmd>[1-3])\((?P<nodeid>[^\)]+)\)(\:(\((?P<claaddr>[^\)]+)\))?"
    r"(\:(?P<eidlist>\[(\(([^\)]+)\),?)*\])?"
    r"(\:(?P<contactlist>\[(\{[0-9]+,[0-9]+,[0-9]+(,\[(\([^\)]+\),?)+\])?\}"
    r",?)*\])?)?)?)?;$"
)
EID_LIST_PARSE_REGEX = re.compile(
    r"^[\[,]\((?P<eid>[^\)]+)\)(?P<remainder>,\(.+\))?\]?$"
)
CONTACT_LIST_PARSE_REGEX = re.compile(
    r"^[\[,]\{(?P<start>[0-9]+),(?P<end>[0-9]+),(?P<datarate>[0-9]+)"
    r"(,(?P<eidlist>\[(\([^\)]+\),?)+\]))?\}(?P<remainder>,\{.+\})?\]?$"
)

# uD3TN config protocol operations
OP_ADD = 1
OP_REPLACE = 2
OP_DELETE = 3

# Which DispatchReason values we accept for new bundles to be scheduled
# (in other cases they are dropped).
VALID_SCHEDULE_DISPATCH_REASONS = (
    DispatchReason.DISPATCH_REASON_NO_FIB_ENTRY,
)

logger = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True)
class ContactDefinition:
    """Represents a planned contact"""
    start: int
    end: int
    next_hop_node_id: str
    next_hop_cla_addr: str
    data_rate: int
    # All following fields are not part of the ID
    reachable_eids: Set[str] = dataclasses.field(compare=False)


@dataclasses.dataclass(frozen=True)
class BundleIDType:
    """Represents a bundle scheduled in the BDM"""
    src_eid: str
    creation_timestamp: int
    sequence_number: int
    fragment_offset: int
    payload_length: int
    # All following fields are not part of the ID
    dst_eid: str = dataclasses.field(compare=False)
    min_frag_size_first: int = dataclasses.field(compare=False)
    min_frag_size_last: int = dataclasses.field(compare=False)

    def to_proto(self):
        return CompoundBundleId(
            source_eid=self.src_eid,
            creation_timestamp=self.creation_timestamp,
            sequence_number=self.sequence_number,
            fragment_offset=self.fragment_offset,
            payload_length=self.payload_length,
        )


SchedDBType = Dict[
    ContactDefinition,
    Tuple[
        asyncio.Queue,
        Dict[BundleIDType, Tuple[int, int, int]],
        asyncio.Task,
    ],
]

FragmentDispatchTupleType = Tuple[ContactDefinition, int, int, int]
FragmentDispatchSetType = Set[FragmentDispatchTupleType]
SchedDictType = Dict[
    BundleIDType,
    FragmentDispatchSetType,
]


class ScheduleException(Exception):
    pass


class FIBLinkStatusManager:

    def __init__(self):
        self.known_links: Dict[str, asyncio.Event] = {}

    def update_link_info(self, next_hop: str, is_active: bool = False):
        # NOTE: ATM `next_hop` is a CLA addr; uD3TN does not always report EIDs
        if next_hop not in self.known_links:
            ev = asyncio.Event()
            self.known_links[next_hop] = ev
        else:
            ev = self.known_links[next_hop]
        if is_active:
            ev.set()
        else:
            ev.clear()

    def is_active(self, next_hop: str) -> bool:
        if next_hop not in self.known_links:
            return False
        return self.known_links[next_hop].is_set()

    async def wait_until_active(self, next_hop: str):
        if next_hop not in self.known_links:
            ev = asyncio.Event()
            self.known_links[next_hop] = ev
        else:
            ev = self.known_links[next_hop]
        await ev.wait()


def _parse_eid_list(eid_list: str) -> Set[str]:
    if eid_list is None:
        return set()
    match_obj = EID_LIST_PARSE_REGEX.match(eid_list)
    if match_obj is None:
        return set()
    if match_obj.group(2) is None:
        return {match_obj.group(1)}
    return {match_obj.group(1)} | _parse_eid_list(match_obj.group(2))


def _parse_contact_list(
    contact_list: str,
    next_hop_node_id: str,
    next_hop_cla_addr: str,
    reachable_eids_at_node: Set[str]
) -> Set[ContactDefinition]:
    if contact_list is None:
        return set()
    match_obj = CONTACT_LIST_PARSE_REGEX.match(contact_list)
    if match_obj is None:
        return set()
    contact = ContactDefinition(
        start=int(match_obj.group("start")),
        end=int(match_obj.group("end")),
        next_hop_node_id=next_hop_node_id,
        next_hop_cla_addr=next_hop_cla_addr,
        data_rate=int(match_obj.group("datarate")),
        reachable_eids=(
            reachable_eids_at_node
            if match_obj.group("eidlist") is None
            else (
                reachable_eids_at_node |
                _parse_eid_list(match_obj.group("eidlist"))
            )
        )
    )
    if match_obj.group("remainder") is None:
        return {contact}
    return {contact} | _parse_contact_list(
        match_obj.group("remainder"),
        next_hop_node_id,
        next_hop_cla_addr,
        reachable_eids_at_node,
    )


def _parse_config(config_str: str) -> Tuple[int, str, Set[ContactDefinition]]:
    match_obj = CONFIG_PARSE_REGEX.match(config_str)
    if match_obj is None:
        raise ValueError("invalid config string")
    return (
        int(match_obj.group("cmd")),
        match_obj.group("nodeid"),
        _parse_contact_list(
            match_obj.group("contactlist"),
            match_obj.group("nodeid"),
            match_obj.group("claaddr"),
            (
                {match_obj.group("nodeid")} |
                _parse_eid_list(match_obj.group("eidlist"))
            ),
        ),
    )


def _contact_overlaps(c1: ContactDefinition, c2: ContactDefinition) -> bool:
    return (
        c1.start < c2.end and
        c1.end > c2.start
    )


def _adjust_existing_contacts(
    global_contact_set: Set[ContactDefinition],
    next_hop_node_id: str,
    contact_set_to_add: Set[ContactDefinition]
):
    # if any of the contacts to add overlaps, remove (-> replace) the existing
    # but extend the "reachable EIDs" set in the new contact to add
    contacts_to_remove = set()
    for contact in global_contact_set:
        # the existing contact is with the same next hop as the ones to add
        if contact.next_hop_node_id == next_hop_node_id:
            for contact2 in contact_set_to_add:
                if contact != contact2 and _contact_overlaps(contact,
                                                             contact2):
                    contacts_to_remove.add(contact)
                    contact2.reachable_eids.update(contact.reachable_eids)
    return contacts_to_remove


def _decode_cfg_msg(
    global_contact_set: Set[ContactDefinition],
    adu_msg: BundleADU,
    payload: bytes,
) -> Tuple[Set[ContactDefinition], Set[ContactDefinition]]:
    operation, nh_node_id, contact_set = _parse_config(
        payload.decode("utf-8")
    )
    contacts_to_add = set()
    contacts_to_remove = set()
    if operation == OP_ADD:
        contacts_to_remove |= _adjust_existing_contacts(
            global_contact_set,
            nh_node_id,
            contact_set,
        )
        # NOTE: does not extend contacts that overlap (ud3tn v0.13 behavior)
        contacts_to_add |= contact_set
    elif operation == OP_REPLACE:
        for contact in global_contact_set:
            # remove all contacts with the node that is to be replaced
            if (contact.next_hop_node_id == nh_node_id and
                    contact not in contact_set):
                contacts_to_remove.add(contact)
        contacts_to_add |= contact_set
    elif operation == OP_DELETE:
        if len(contact_set) == 0:
            # delete all
            for contact in global_contact_set:
                if contact.next_hop_node_id == nh_node_id:
                    contacts_to_remove.add(contact)
        else:
            # delete specified contacts
            contacts_to_remove |= contact_set
    return contacts_to_add, contacts_to_remove


def _rebuild_fwd_dict(
    contact_set: Set[ContactDefinition]
) -> Dict[str, Set[ContactDefinition]]:
    fwd_dict: Dict[str, Set[ContactDefinition]] = {}
    for contact in contact_set:
        for reachable_node_id in contact.reachable_eids:
            try:
                normalized_node_id = get_node_id(reachable_node_id)
            except ValueError:
                logger.info(
                    "Cannot get node ID for EID \"%s\", skipping",
                    reachable_node_id,
                )
                continue
            if normalized_node_id not in fwd_dict:
                fwd_dict[normalized_node_id] = set()
            fwd_dict[normalized_node_id].add(contact)
    return fwd_dict


def _loop_time_to_contact_time(lt: float) -> int:
    return round(lt - UNIX_TO_DTN_OFFSET)


def _contact_time_to_loop_time(ct: int) -> int:
    return ct + UNIX_TO_DTN_OFFSET


def _get_cur_remaining_contact_capacity(
    c: ContactDefinition,
    t: float,
    capmap: Dict[ContactDefinition, Dict[BundleIDType, int]]
) -> int:
    ct = _loop_time_to_contact_time(t)
    if ct >= c.end:
        return 0
    if ct > c.start:
        total_cap = (c.end - ct) * c.data_rate
    else:
        total_cap = (c.end - c.start) * c.data_rate
    try:
        scheduled_cap = sum(capmap[c].values())
    except KeyError:
        scheduled_cap = 0
    # return max(0, total_cap - scheduled_cap)
    return int(total_cap - scheduled_cap)


# NOTE this function is also used for re-scheduling (only in this case,
# fragment_offset and fragment_length are set).
def _calc_bundle_schedule(
    fwd_dict: Dict[str, Set[ContactDefinition]],
    capmap: Dict[ContactDefinition, Dict[BundleIDType, int]],
    bid: BundleIDType,
    mbs_bytes: int,
    # allow to override the following for re-scheduling
    fragment_offset: int = 0,
    fragment_length: int = 0,
) -> FragmentDispatchSetType:
    # Main dispatch logic for a new bundle
    try:
        dst_node_id = get_node_id(bid.dst_eid)
    except ValueError:
        raise ScheduleException("cannot determine next-hop node ID")
    next_hop_contacts = fwd_dict.get(dst_node_id, None)
    if not next_hop_contacts:
        raise ScheduleException("no next hops found")
    # Try to schedule
    bundle_first_fragment_min_size_bytes = bid.min_frag_size_first
    bundle_last_fragment_min_size_bytes = bid.min_frag_size_last
    frag_remaining_payload = (
        bid.payload_length if fragment_length == 0 else fragment_length
    )
    if fragment_offset == 0:
        fragment_offset = bid.fragment_offset
    logger.debug(
        "Bundle fragmentation parameters: first_frag_min = %d, "
        "last_frag_min = %d, pl_sz = %d",
        bundle_first_fragment_min_size_bytes,
        bundle_last_fragment_min_size_bytes,
        frag_remaining_payload,
    )
    fragment_results: List[FragmentDispatchTupleType] = []
    for contact in sorted(next_hop_contacts, key=lambda c: c.start):
        cur_time = time.time()
        # Determine max. bundle size for this contact
        contact_crcap_bytes: int = _get_cur_remaining_contact_capacity(
            contact,
            cur_time,
            capmap,
        )
        if mbs_bytes:
            contact_mbs_bytes = min(contact_crcap_bytes, mbs_bytes)
        else:
            contact_mbs_bytes = contact_crcap_bytes
        if bundle_last_fragment_min_size_bytes == 0:
            # "Must not fragment" bit is set
            bundle_max_bytes: int = bundle_first_fragment_min_size_bytes
            bundle_header_size: int = 0
            bundle_min_size: int = bundle_first_fragment_min_size_bytes
        else:
            bundle_fragment_min_size_bytes = (
                bundle_first_fragment_min_size_bytes
                if not fragment_results
                else bundle_last_fragment_min_size_bytes
            )
            bundle_max_bytes = (
                bundle_fragment_min_size_bytes + frag_remaining_payload
            )
            bundle_header_size = bundle_fragment_min_size_bytes
            # At least the headers + one byte of the payload must be
            # transferred, except if the bundle must not be fragmented.
            bundle_min_size = bundle_fragment_min_size_bytes + 1
        logger.debug(
            "Considering contact: %s (cap = %d, bms = %d, mbs = %d, now = %d)",
            contact,
            contact_crcap_bytes,
            bundle_min_size,
            contact_mbs_bytes,
            _loop_time_to_contact_time(cur_time),
        )
        if contact_mbs_bytes < bundle_min_size:
            continue
        fragment_size: int = min(bundle_max_bytes, contact_mbs_bytes)
        fragment_payload_size = (
            fragment_size - bundle_header_size
        )
        logger.debug(
            "Assigning fragment of size = %d (pl_sz = %d)",
            fragment_size,
            fragment_payload_size,
        )
        if fragment_payload_size == bid.payload_length:
            # push the complete bundle
            fragment_results.append((
                contact,
                bid.fragment_offset,
                bid.payload_length,
                fragment_size,
            ))
        else:
            fragment_results.append((
                contact,
                fragment_offset,
                fragment_payload_size,
                fragment_size,
            ))
        fragment_offset += fragment_payload_size
        frag_remaining_payload -= fragment_payload_size
        if frag_remaining_payload <= 0:
            break
    if frag_remaining_payload > 0:
        raise ScheduleException("insufficient contact volume available")
    return set(fragment_results)


def _bundle_id(bdl: Bundle, bdl_disp_info: BundleDispatchInfo):
    return BundleIDType(
        src_eid=bdl.src_eid,
        creation_timestamp=bdl.creation_timestamp_ms,
        sequence_number=bdl.sequence_number,
        fragment_offset=bdl.fragment_offset,
        payload_length=bdl.payload_length,
        dst_eid=bdl.dst_eid,
        min_frag_size_first=bdl_disp_info.min_frag_size_first,
        min_frag_size_last=bdl_disp_info.min_frag_size_last,
    )


async def sleep_until(simtime: float) -> None:
    """Sleep using asyncio.sleep until the specified time is reached.

    This function returns immediately if the specified timestamp is in the
    past. This function may only be called from within asyncio callbacks,
    tasks, or coroutines.
    """
    duration = simtime - time.time()
    if duration > 0:
        await asyncio.sleep(duration)


def _build_storage_call(op: StorageOperation, bid: BundleIDType) -> bytes:
    cmd = StorageCall(
        operation=op,
        id=bid.to_proto(),
    )
    cmd_bytes = cmd.SerializeToString()
    return proto_encoder._VarintBytes(len(cmd_bytes)) + cmd_bytes


async def _send_storage_call(
    aap2_rpc_client: AAP2AsyncClient,
    storage_agent_eid: str,
    op: StorageOperation,
    bid: BundleIDType
):
    payload = _build_storage_call(
        op,
        bid,
    )
    assert (
        (await aap2_rpc_client.send_adu_rpc(
            BundleADU(
                dst_eid=storage_agent_eid,
                payload_length=len(payload),
                adu_flags=[BundleADUFlags.BUNDLE_ADU_WITH_BDM_AUTH],
            ),
            payload,
        )).response_status ==
        ResponseStatus.RESPONSE_STATUS_SUCCESS
    )


async def _push_bundle_from_storage(
    aap2_rpc_client: AAP2AsyncClient,
    storage_agent_eid: str,
    bid: BundleIDType
):
    await _send_storage_call(
        aap2_rpc_client,
        storage_agent_eid,
        StorageOperation.STORAGE_OPERATION_PUSH_BUNDLES,
        bid,
    )


async def _delete_bundle_from_storage(
    aap2_rpc_client: AAP2AsyncClient,
    storage_agent_eid: str,
    bid: BundleIDType
):
    await _send_storage_call(
        aap2_rpc_client,
        storage_agent_eid,
        StorageOperation.STORAGE_OPERATION_DELETE_BUNDLES,
        bid,
    )


async def _bundle_transmission(
    aap2_rpc_client: AAP2AsyncClient,
    storage_agent_eid: str,
    contact: ContactDefinition,
    queue: asyncio.Queue
):
    while True:
        try:
            bid = await queue.get()
        except (asyncio.IncompleteReadError, asyncio.CancelledError):
            return
        # NOTE: if the bundle is not in storage anymore, e.g., because it was
        # dropped or re-scheduled, this just silently fails. In the future we
        # may want to provide functionality to inform the BDM of that failure
        # (e.g., by a response sent by the storage agent).
        try:
            logger.debug(
                "Pushing bundle %s during contact %s",
                bid,
                contact,
            )
            await _push_bundle_from_storage(
                aap2_rpc_client,
                storage_agent_eid,
                bid,
            )
        except (asyncio.IncompleteReadError, asyncio.CancelledError):
            logger.debug(
                "Re-inserting bundle %s after termination of contact %s",
                bid,
                contact,
            )
            queue.put_nowait(bid)  # ensure it is re-scheduled
            return
        except Exception as e:
            logger.warning(
                "Bundle TX terminated with %s: %s",
                type(e).__name__,
                e,
            )
            raise


async def _reschedule_bundles(
    aap2_rpc_client: AAP2AsyncClient,
    storage_agent_eid: str,
    fwd_dict: Dict[str, Set[ContactDefinition]],
    capmap: Dict[ContactDefinition, Dict[BundleIDType, int]],
    scheduled_bundles: SchedDictType,
    sched_db: SchedDBType,
    contact: ContactDefinition,
):
    queue, sched_set, _ = sched_db[contact]
    # 1. get all bundle IDs from the scheduling queue for the contact
    # Note that every bundle may be scheduled just in part for that contact,
    # which means that parts (other fragments) are scheduled for other contacts
    # or were already sent.
    bid_set = set()
    try:
        while True:
            bid_set.add(queue.get_nowait())
    except asyncio.QueueEmpty:
        pass
    # 2. attempt to schedule each bid again
    for bid in bid_set:
        fo, fps, fs = sched_set[bid]
        try:
            fragment_results = _calc_bundle_schedule(
                fwd_dict,
                capmap,
                bid,
                0,  # no mbs as we already scheduled this bundle appropriately
                # from sched_set, as ID is for the whole/as-received bundle!
                fo,
                fps,
            )
            if len(fragment_results) == 0:
                raise ScheduleException("no route found")
        except ScheduleException as e:
            # any part failed -> drop it everywhere
            # drop from storage -> other parts may be pushed but do not arrive
            logger.info(
                "Could not re-schedule Bundle %s, deleting: %s",
                bid,
                e,
            )
            await _delete_bundle_from_storage(
                aap2_rpc_client,
                storage_agent_eid,
                bid,
            )
        else:
            logger.debug(
                "Bundle %s frag=(%d, %d) re-scheduled for: %s",
                bid,
                fo,
                fps,
                fragment_results,
            )
            for c, c_fo, c_fps, c_fs in fragment_results:
                c_queue, c_sched_set, _ = sched_db[c]
                c_queue.put_nowait(bid)
                c_sched_set[bid] = (c_fo, c_fps, c_fs)
            scheduled_bundles[bid].update(fragment_results)


async def _schedule_contact(
    aap2_rpc_client: AAP2AsyncClient,
    storage_agent_eid: str,
    active_contacts: Set[ContactDefinition],
    contact: ContactDefinition,
    contact_queue: asyncio.Queue,
    flsm: FIBLinkStatusManager,
    fwd_dict: Dict[str, Set[ContactDefinition]],
    capmap: Dict[ContactDefinition, Dict[BundleIDType, int]],
    scheduled_bundles: SchedDictType,
    sched_db: SchedDBType,
):
    start_time_loop = _contact_time_to_loop_time(contact.start)
    end_time_loop = _contact_time_to_loop_time(contact.end)
    logger.debug(
        "Contact: %s - sleeping until %d (%d s)",
        contact,
        start_time_loop,
        start_time_loop - time.time()
    )
    try:
        await sleep_until(start_time_loop)
        logger.debug("Contact started: %s", contact)
        active_contacts.add(contact)
        # initiate link via CLA
        assert (
            (await aap2_rpc_client.send_rpc(AAPMessage(link=Link(
                status=LinkStatus.LINK_STATUS_UP,
                peer_node_id=contact.next_hop_node_id,
                peer_cla_addr=contact.next_hop_cla_addr,
            )))).response_status ==
            ResponseStatus.RESPONSE_STATUS_SUCCESS
        )
        # Wait for link to become active if it is not already...
        await flsm.wait_until_active(contact.next_hop_cla_addr)
        try:
            await asyncio.wait_for(
                _bundle_transmission(
                    aap2_rpc_client,
                    storage_agent_eid,
                    contact,
                    contact_queue,
                ),
                timeout=(end_time_loop - time.time()),
            )
        except asyncio.TimeoutError:
            pass
        except Exception as e:
            logger.warning(
                "Bundle transmission failed for contact: %s; %s: %s",
                contact,
                type(e).__name__,
                e,
            )
            raise
        else:
            logger.debug(
                "Bundle transmission stopped for contact: %s",
                contact,
            )
            await sleep_until(end_time_loop)
    except asyncio.CancelledError:
        pass
    finally:
        logger.debug("Contact ended or terminated: %s", contact)
        await _reschedule_bundles(
            aap2_rpc_client,
            storage_agent_eid,
            fwd_dict,
            capmap,
            scheduled_bundles,
            sched_db,
            contact,
        )
        # in any case, attempt to tear down the CLA link
        assert (
            (await aap2_rpc_client.send_rpc(AAPMessage(link=Link(
                status=LinkStatus.LINK_STATUS_DOWN,
                peer_node_id=contact.next_hop_node_id,
                peer_cla_addr=contact.next_hop_cla_addr,
            )))).response_status ==
            ResponseStatus.RESPONSE_STATUS_SUCCESS
        )
        try:
            active_contacts.remove(contact)
            del sched_db[contact]
        except KeyError:
            pass


async def _dispatch_scheduled_bundle(
    scheduled_bundles: SchedDictType,
    sched_db: SchedDBType,
    active_contacts: Set[ContactDefinition],
    bid: BundleIDType
) -> Tuple[DispatchResult, FragmentDispatchSetType]:
    # Extract the set next hop + fragmentation settings from the
    # active contacts and then dispatch it there...
    fragment_dispatch_commands = set()
    for contact, _, _, _ in scheduled_bundles[bid]:
        if contact in active_contacts:
            try:
                queue, sched_set, task = sched_db[contact]
            except KeyError:
                continue
            fo, fps, fs = sched_set[bid]
            fragment_dispatch_commands.add(
                (contact, fo, fps, fs)
            )
    # If the length of the `fragment_dispatch_commands` set is 0,
    # the bundle should not be forwarded right now (or not anymore,
    # as rescheduling failed?).
    # Might be that the fuzzy EID glob called a wrong bundle
    # from storage. We explicitly drop it from the working
    # memory (forward to the empty set) as we can just call it
    # again later.
    disp_result = DispatchResult(
        next_hops=[
            DispatchResult.NextHopEntry(
                node_id=contact.next_hop_node_id,
                fragment_offset=fo,
                fragment_length=fs,
            )
            for contact, fo, fs, _ in fragment_dispatch_commands
        ],
    )
    logger.debug(
        "Scheduling entry found, dispatching result: %s",
        disp_result,
    )
    return disp_result, fragment_dispatch_commands


def _schedule_bundle(
    bid: BundleIDType,
    fwd_dict: Dict[str, Set[ContactDefinition]],
    capmap: Dict[ContactDefinition, Dict[BundleIDType, int]],
    scheduled_bundles: SchedDictType,
    sched_db: SchedDBType,
    storage_node_id: str,
    mbs_bytes: int,
) -> DispatchResult:
    try:
        fragment_results = _calc_bundle_schedule(
            fwd_dict,
            capmap,
            bid,
            mbs_bytes,
        )
        if len(fragment_results) == 0:
            raise ScheduleException("no route found")
    except ScheduleException as e:
        logger.debug(
            "Could not schedule bundle, issuing empty dispatch: %s",
            e,
        )
        # Drop bundle if we cannot schedule it (v0.13 behavior)
        # Dropping means generating the dispatch response worked
        # (SUCCESS), but we want to dispatch to the empty set.
        return DispatchResult(next_hops=[])
    except Exception as e:
        logger.warning("Scheduling failed: %s", e)
        raise
    else:
        scheduled_bundles[bid] = fragment_results
        for contact, _, _, fs in fragment_results:
            if contact not in capmap:
                capmap[contact] = {}
            capmap[contact][bid] = fs
        logger.debug(
            "Bundle %s scheduled for: %s",
            bid,
            fragment_results,
        )
        # Always dispatch to storage when scheduled successfully.
        return DispatchResult(
            next_hops=[DispatchResult.NextHopEntry(node_id=storage_node_id)],
        )


async def _dispatch_bundle(
    bid: BundleIDType,
    active_contacts: Set[ContactDefinition],
    scheduled_bundles: SchedDictType,
    sched_db: SchedDBType,
    stored_bundles: Set[BundleIDType],
    aap2_rpc_client: AAP2AsyncClient,
    storage_agent_eid: str,
    prevent_loops: bool,
) -> DispatchResult:
    if bid not in stored_bundles:
        logger.debug(
            "Bundle %s not stored (dropped on re-scheduling?), ignoring",
            bid,
        )
        return None
    # If no TX event, it is something we recalled from storage.
    dispatch_result, frag_dispatched = await _dispatch_scheduled_bundle(
        scheduled_bundles,
        sched_db,
        active_contacts,
        bid,
    )
    scheduled_bundles[bid] -= frag_dispatched
    # Delete the bundle from storage - a copy should be in the BP agent
    # connection still, until it has been transmitted finally.
    if len(scheduled_bundles[bid]) == 0:
        # NOTE: This emulates the v0.13.0 behavior - we always
        # drop a bundle after *attempting* to send it.
        logger.debug(
            "Bundle %s sent, deleting from storage.",
            str(bid),
        )
        await _delete_bundle_from_storage(
            aap2_rpc_client,
            storage_agent_eid,
            bid,
        )
        # NOTE if at some point we attempt re-dispatch, e.g., because of
        # failed TX, we must properly reference-count transmissions or
        # always keep bundle references here.
        if not prevent_loops:
            del scheduled_bundles[bid]
            stored_bundles.discard(bid)
    return dispatch_result


async def _process_dispatch_event(
    dr: DispatchEvent,
    fwd_dict: Dict[str, Set[ContactDefinition]],
    capmap: Dict[ContactDefinition, Dict[BundleIDType, int]],
    scheduled_bundles: SchedDictType,
    sched_db: SchedDBType,
    stored_bundles: Set[BundleIDType],
    active_contacts: Set[ContactDefinition],
    aap2_rpc_client: AAP2AsyncClient,
    storage_agent_eid: str,
    storage_node_id: str,
    prevent_loops: bool,
) -> DispatchResult:
    bdl = dr.bundle
    bid = _bundle_id(bdl, dr.additional_information)
    reason = dr.reason
    logger.debug(
        "Received dispatch request with reason %s for bundle %s to %s",
        str(reason),
        bid,
        bdl.dst_eid,
    )
    if bid in scheduled_bundles:
        if reason == DispatchReason.DISPATCH_REASON_TX_SUCCEEDED:
            if bid not in stored_bundles:
                # This was the initial "TX to storage"
                fragment_results = scheduled_bundles[bid]
                for contact, fo, fps, fs in fragment_results:
                    queue, sched_set, task = sched_db[contact]
                    queue.put_nowait(bid)
                    if bid in sched_set:
                        logger.info(
                            "Looping bundle recognized in active contact: %s",
                            bid,
                        )
                    sched_set[bid] = (fo, fps, fs)
                stored_bundles.add(bid)
        elif reason == DispatchReason.DISPATCH_REASON_TX_FAILED:
            # NOTE: Same behavior as v0.13.0 - this is the compat. BDM.
            logger.warning(
                "TX failed for bundle %s, DROPPING.",
                str(bid),
            )
        else:
            return await _dispatch_bundle(
                bid,
                active_contacts,
                scheduled_bundles,
                sched_db,
                stored_bundles,
                aap2_rpc_client,
                storage_agent_eid,
                prevent_loops,
            )
    elif reason not in VALID_SCHEDULE_DISPATCH_REASONS:
        logger.debug(
            "No scheduling entry found, but dispatch reason is "
            "%s, thus, ignoring it",
            str(reason),
        )
    else:
        return _schedule_bundle(
            bid,
            fwd_dict,
            capmap,
            scheduled_bundles,
            sched_db,
            storage_node_id,
            dr.additional_information.max_bundle_size_bytes,
        )
    return None


def _process_config(
    aap2_rpc_client: AAP2AsyncClient,
    adu: BundleADU,
    payload: bytes,
    contact_set: Set[ContactDefinition],
    fwd_dict: Dict[str, Set[ContactDefinition]],
    capmap: Dict[ContactDefinition, Dict[BundleIDType, int]],
    scheduled_bundles: SchedDictType,
    sched_db: SchedDBType,
    active_contacts: Set[ContactDefinition],
    flsm: FIBLinkStatusManager,
    storage_agent_eid: str,
    insecure_config: bool,
) -> Dict[str, Set[ContactDefinition]]:
    logger.debug(
        "Processing new configuration message from %s",
        adu.src_eid,
    )
    if (BundleADUFlags.BUNDLE_ADU_WITH_BDM_AUTH not in adu.adu_flags
            and not insecure_config):
        logger.warning(
            "Flag `BUNDLE_ADU_WITH_BDM_AUTH` not set for received config "
            "message and `--insecure-config` argument is not set, thus, "
            "dropping message as we cannot validate it is secure"
        )
        return fwd_dict
    add_set, rm_set = _decode_cfg_msg(
        contact_set,
        adu,
        payload,
    )
    contact_set.difference_update(rm_set)
    for contact in rm_set:
        logger.debug("Removing contact: %s", contact)
        try:
            queue, sched_set, task = sched_db[contact]
        except KeyError:
            continue
        task.cancel()
        # Will delete itself!
        # del sched_db[contact]
    contact_set.update(add_set)
    for contact in add_set:
        logger.debug("Adding contact: %s", contact)
        contact_queue: asyncio.Queue[BundleIDType] = asyncio.Queue()
        task = asyncio.create_task(_schedule_contact(
            aap2_rpc_client,
            storage_agent_eid,
            active_contacts,
            contact,
            contact_queue,
            flsm,
            # re-scheduling
            fwd_dict,
            capmap,
            scheduled_bundles,
            sched_db,
        ))
        assert contact not in sched_db
        sched_db[contact] = (contact_queue, {}, task)
    return _rebuild_fwd_dict(contact_set)


def _update_link_info(flsm: FIBLinkStatusManager, link: Link):
    next_hop = link.peer_cla_addr
    is_now_active = (link.status == LinkStatus.LINK_STATUS_UP)
    flsm.update_link_info(next_hop, is_now_active)
    logger.debug(
        "Updating link status for link with '%s' via '%s' to %d",
        link.peer_node_id,
        link.peer_cla_addr,
        is_now_active,
    )


async def run_compat_bdm(
    aap2_sub_client: AAP2AsyncClient,
    aap2_rpc_client: AAP2AsyncClient,
    storage_agent_eid: str,
    storage_node_id: str,
    prevent_loops: bool,
    insecure_config: bool,
):
    contact_set: Set[ContactDefinition] = set()
    active_contacts: Set[ContactDefinition] = set()
    # scheduled_bundles: Dict[BundleIDType, Set[ContactDefinition]] = {}
    scheduled_bundles: SchedDictType = {}
    capmap: Dict[ContactDefinition, Dict[BundleIDType, int]] = {}
    stored_bundles: Set[BundleIDType] = set()
    fwd_dict: Dict[str, Set[ContactDefinition]] = {}
    sched_db: SchedDBType = {}
    flsm = FIBLinkStatusManager()
    logger.info("Waiting for dispatch requests...")

    while True:
        try:
            msg = await aap2_sub_client.receive_msg()
        except AAP2ServerDisconnected:
            logger.warning("µD3TN has closed the connection.")
            sys.exit(1)
        except asyncio.exceptions.CancelledError:
            logger.info("Terminated by keyboard interrupt.")
            sys.exit(130)  # exit status for SIGINT
        response_status = ResponseStatus.RESPONSE_STATUS_SUCCESS
        dispatch_result = None
        if msg.WhichOneof("msg") == "dispatch_event":
            dispatch_result = await _process_dispatch_event(
                msg.dispatch_event,
                fwd_dict,
                capmap,
                scheduled_bundles,
                sched_db,
                stored_bundles,
                active_contacts,
                aap2_rpc_client,
                storage_agent_eid,
                storage_node_id,
                prevent_loops,
            )
        elif msg.WhichOneof("msg") == "adu":
            _, payload = await aap2_sub_client.receive_adu(msg.adu)
            fwd_dict_new = _process_config(
                aap2_rpc_client,
                msg.adu,
                payload,
                contact_set,
                fwd_dict,
                capmap,
                scheduled_bundles,
                sched_db,
                active_contacts,
                flsm,
                storage_agent_eid,
                insecure_config,
            )
            # Update in-place as contact tasks may hold a reference!
            fwd_dict.clear()
            fwd_dict.update(fwd_dict_new)
        elif msg.WhichOneof("msg") == "link":
            _update_link_info(flsm, msg.link)
        elif msg.WhichOneof("msg") == "keepalive":
            response_status = ResponseStatus.RESPONSE_STATUS_ACK
        else:
            logger.debug("Received message with field '%s' set, discarding.",
                         msg.WhichOneof("msg"))
        await aap2_sub_client.send(AAPResponse(
            response_status=response_status,
            dispatch_result=dispatch_result,
        ))


async def execute_bdm(
    aap2_sub_client: AAP2AsyncClient,
    aap2_rpc_client: AAP2AsyncClient,
    agentid: str,
    secret: str,
    storage_agent_eid: Optional[str],
    storage_node_id: str,
    prevent_loops: bool,
    insecure_config: bool,
    keepalive_seconds: int,
):
    async with aap2_sub_client, aap2_rpc_client:
        is_ipn = aap2_sub_client.is_ipn_eid
        default_config_agent_id = (
            DEFAULT_CONFIG_AGENT_ID_IPN
            if is_ipn
            else DEFAULT_CONFIG_AGENT_ID_DTN
        )
        secret = await aap2_sub_client.configure(
            agentid or default_config_agent_id,
            subscribe=True,
            secret=secret,
            auth_type=AuthType.AUTH_TYPE_FIB_AND_DISPATCH,
            keepalive_seconds=keepalive_seconds,
        )
        await aap2_rpc_client.configure(
            agentid or default_config_agent_id,
            subscribe=False,
            secret=secret,
            auth_type=AuthType.AUTH_TYPE_FIB_AND_DISPATCH,
        )
        if storage_agent_eid is None:
            if aap2_rpc_client.is_ipn_eid:
                storage_agent_eid = f"{aap2_rpc_client.eid_prefix}.9003"
            else:
                storage_agent_eid = f"{aap2_rpc_client.eid_prefix}sqlite"
        logger.info("AAP 2.0 agents registered and configured!")
        logger.info("Assigned agent secret: '%s'", secret)
        await run_compat_bdm(
            aap2_sub_client,
            aap2_rpc_client,
            storage_agent_eid,
            storage_node_id,
            prevent_loops,
            insecure_config,
        )


def main():
    parser = argparse.ArgumentParser(
        description=(
            "bundle dispatcher module realizing the uD3TN v0.13 forwarding"
            "approach via the AAP 2.0 BDM interface"
        ),
    )
    add_common_parser_arguments(parser)
    add_keepalive_parser_argument(parser)
    parser.add_argument(
        "--storage-agent-eid",
        default=None,
        help="the (full) EID of the bundle storage agent (default: <auto>)",
    )
    parser.add_argument(
        "--storage-node-id",
        default="dtn:storage",
        help="the next-hop node ID of the bundle storage system",
    )
    parser.add_argument(
        "--prevent-loops",
        action="store_true",
        help="keep bundle references, so we find and drop looping bundles",
    )
    parser.add_argument(
        "--insecure-config",
        action="store_true",
        help=(
            "do not check that the sender of a config bundle is local and has "
            "authenticated with the administrative / BDM credential"
        ),
    )

    args = parser.parse_args()
    global logger
    logger = initialize_logger(args.verbosity)

    if args.tcp:
        aap2_sub_client = AAP2AsyncTCPClient(address=args.tcp)
        aap2_rpc_client = AAP2AsyncTCPClient(address=args.tcp)
    else:
        aap2_sub_client = AAP2AsyncUnixClient(address=args.socket)
        aap2_rpc_client = AAP2AsyncUnixClient(address=args.socket)
    asyncio.run(execute_bdm(
        aap2_sub_client,
        aap2_rpc_client,
        args.agentid,
        get_secret_from_args(args),
        args.storage_agent_eid,
        args.storage_node_id,
        args.prevent_loops,
        args.insecure_config,
        args.keepalive_seconds,
    ))


if __name__ == "__main__":
    main()


def test_parse_config():
    assert _parse_config(
        "1(dtn://ud3tn2.dtn/):(mtcp:127.0.0.1:4223)::["
        "{1401519306972,1401519316972,1200,[(dtn://89326/),(dtn://12349/)]},"
        "{1401519506972,1401519516972,1200,[(dtn://89326/),(dtn://12349/)]}];"
    ) == (
        1,
        "dtn://ud3tn2.dtn/",
        {
            ContactDefinition(
                start=1401519306972,
                end=1401519316972,
                data_rate=1200,
                next_hop_node_id="dtn://ud3tn2.dtn/",
                next_hop_cla_addr="mtcp:127.0.0.1:4223",
                reachable_eids={
                    "dtn://89326/",
                    "dtn://12349/",
                },
            ),
            ContactDefinition(
                start=1401519506972,
                end=1401519516972,
                data_rate=1200,
                next_hop_node_id="dtn://ud3tn2.dtn/",
                next_hop_cla_addr="mtcp:127.0.0.1:4223",
                reachable_eids={
                    "dtn://89326/",
                    "dtn://12349/",
                },
            ),
        },
    )
    assert _parse_config(
        "1(dtn://ud3tn2.dtn/)::[(dtn://18471/),(dtn://81491/)]:"
        "[{1401519406972,1401819306972,1200}];"
    ) == (
        1,
        "dtn://ud3tn2.dtn/",
        {
            ContactDefinition(
                start=1401519406972,
                end=1401819306972,
                data_rate=1200,
                next_hop_node_id="dtn://ud3tn2.dtn/",
                next_hop_cla_addr=None,
                reachable_eids=[
                    "dtn://18471/",
                    "dtn://81491/",
                ],
            ),
        },
    )
    assert _parse_config(
        "2(dtn://ud3tn2.dtn/):(mtcp:127.0.0.1:4223):"
        "[(dtn://89326/),(dtn://12349/)];"
    ) == (
        2,
        "dtn://ud3tn2.dtn/",
        set(),
    )
    assert _parse_config(
        "2(dtn://ud3tn2.dtn/)::[]:[];"
    ) == (
        2,
        "dtn://ud3tn2.dtn/",
        set(),
    )
    assert _parse_config(
        "2(ipn:1.0)::[];"
    ) == (
        2,
        "ipn:1.0",
        set(),
    )
    assert _parse_config(
        "3(dtn://ud3tn2.dtn/);"
    ) == (
        3,
        "dtn://ud3tn2.dtn/",
        set(),
    )
    assert _parse_config(
        "1(dtn://13714/):(tcpspp:):[(dtn://18471/),(dtn://81491/)]:;"
    ) == (
        1,
        "dtn://13714/",
        set(),
    )



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause OR Apache-2.0
# encoding: utf-8

import argparse
import logging

from ud3tn_utils.aap import AAPTCPClient, AAPUnixClient
from ud3tn_utils.aap.bin.helpers import (
    add_socket_group_parser_arguments,
    add_verbosity_parser_argument,
    initialize_logger,
    get_config_eid,
)
from ud3tn_utils.config import ConfigMessage, make_contact


logger = logging.getLogger(__name__)


def main():
    import sys

    parser = argparse.ArgumentParser(
        description="create or update a node in uD3TN",
    )

    add_socket_group_parser_arguments(parser)
    add_verbosity_parser_argument(parser)

    parser.add_argument(
        "--dest_eid",
        default=None,
        help="the EID of the node to which the configuration belongs",
    )
    parser.add_argument(
        "eid",
        help="the EID of the node to which the contact exists",
    )
    parser.add_argument(
        "cla_address",
        help="the CLA address of the node",
    )
    parser.add_argument(
        "-s", "--schedule",
        nargs=3,
        type=int,
        metavar=("start_offset", "duration", "bitrate"),
        action="append",
        default=[],
        help="schedule a contact relative to the current time",
    )
    parser.add_argument(
        "-r", "--reaches",
        type=str,
        action="append",
        default=[],
        help="specify an EID reachable via the node",
    )

    args = parser.parse_args()
    global logger
    logger = initialize_logger(args.verbosity)

    logger.warning(
        "DEPRECATED: This tool does not support new security mechanisms "
        "provided by µD3TN and, thus, has been deprecated. Please use "
        "aap2-config to configure contacts with the compatibility BDM "
        "and provide the administrative secret to both µD3TN and the "
        "configuration tool in release builds."
    )

    if not args.schedule:
        logger.fatal("At least one -s/--schedule argument must be given!")
        sys.exit(1)

    msg = bytes(ConfigMessage(
        args.eid,
        args.cla_address,
        contacts=[
            make_contact(*contact)
            for contact in args.schedule
        ],
        reachable_eids=args.reaches,
    ))

    logger.info("> %s", msg)

    if args.tcp:
        with AAPTCPClient(address=args.tcp) as aap_client:
            aap_client.register()
            dest_eid = args.dest_eid or aap_client.node_eid
            aap_client.send_bundle(get_config_eid(dest_eid), msg)
    else:
        with AAPUnixClient(address=args.socket) as aap_client:
            aap_client.register()
            dest_eid = args.dest_eid or aap_client.node_eid
            aap_client.send_bundle(get_config_eid(dest_eid), msg)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause OR Apache-2.0
# encoding: utf-8

import abc
import asyncio
import logging
import random
import socket
import uuid

from google.protobuf.internal.decoder import _DecodeVarint32
from google.protobuf.internal.encoder import _VarintBytes

from .generated import aap2_pb2


logger = logging.getLogger(__name__)


class AAP2Error(RuntimeError):
    """The base class for errors raised by AAP2Client."""

    pass


class AAP2CommunicationError(AAP2Error):
    """An error raised by AAP2Client when there is a communication error."""

    pass


class AAP2ServerDisconnected(AAP2CommunicationError):
    """An error raised by AAP2Client if the server disconnected."""

    pass


class AAP2OperationFailed(AAP2Error):
    """An error raised by AAP2Client when uD3TN indicates a failure."""

    pass


class AAP2UnexpectedMessage(AAP2Error):
    """An error raised by AAP2Client when receiving an unexpected message."""

    def __init__(self, *args, message=None):
        super().__init__(*args)
        self.message = message


class AAP2Client(abc.ABC):
    """A context manager class for connecting to uD3TN's AAP 2.0 socket.

    Note:
        This is an abstract base class for four concrete implementations:
        `AAP2UnixClient` for synchronous connections to UNIX domain sockets,
        `AAP2TCPClient` for synchronous connections to TCP sockets,
        `AAP2AsyncUnixClient` for asynchronous connections to UNIX domain
        sockets, and `AAP2AsyncTCPClient` for asynchronous connections to
        TCP sockets.

    Args:
        address: The address of the remote socket to connect to.

    Attributes:
        address: The address used to connect to the socket.
        node_eid (str): The local node ID of the connected uD3TN instance.
        agent_id (str): The agent ID assigned to this client.

    """

    def __init__(self, address):
        self.address = address
        self.node_eid = None
        self.agent_id = None

    @property
    def eid(self):
        """Return the own EID including agent ID."""
        if self.node_eid[0:3] == "ipn":
            prefix, _ = self.node_eid.split(".")
            return f"{prefix}.{self.agent_id}"
        else:
            return f"{self.node_eid}{self.agent_id}"

    @property
    def eid_prefix(self):
        """Return the EID prefix."""
        if self.node_eid[0:3] == "ipn":
            prefix, _ = self.node_eid.split(".")
            return prefix
        else:
            return self.node_eid

    @property
    def is_ipn_eid(self):
        """Return True if the connected node uses the `ipn` EID scheme."""
        return self.node_eid[0:3] == "ipn"

    def response_status_name(self, status):
        """Get the name for the specified AAP2 ResponseStatus code.

        Args:
            status: The integer ResponseStatus code returned by µD3TN.

        Return:
            A string representation of status, or "<unknown>" for values not
            part of the ResponseStatus enum.
        """
        return {
            v: n for n, v in aap2_pb2.ResponseStatus.items()
        }.get(
            status,
            "<unknown>",
        )

    def _generate_agent_id(self):
        if not self.node_eid:
            return None
        if self.node_eid[0:3] == "dtn":
            return str(uuid.uuid4())
        elif self.node_eid[0:3] == "ipn":
            return str(
                random.randint(1, 4294967295)
            )  # UINT32_MAX for maximum compatibility
        return None

    def _check_first_byte(self, data):
        """Check first byte received from the server, which must be 0x2f."""
        if not data:
            raise AAP2CommunicationError("Connection broke during handshake")
        if data[0] != 0x2f:
            raise AAP2CommunicationError(
                "Did not receive AAP 2.0 magic number 0x2F, but: " +
                hex(data[0])
            )

    def _process_welcome(self, msg_welcome):
        """Decode and store properties received in the WELCOME message."""
        if msg_welcome.WhichOneof("msg") != "welcome":
            raise AAP2UnexpectedMessage(
                "Expected the 'welcome' oneof field but received: " +
                msg_welcome.WhichOneof("msg"),
                msg_welcome
            )
        self.node_eid = msg_welcome.welcome.node_id
        logger.debug(f"WELCOME message received! ~ EID = {self.node_eid}")
        return msg_welcome.welcome

    def _prepare_config_msg(self, agent_id=None, subscribe=False, secret=None,
                            auth_type=aap2_pb2.AuthType.AUTH_TYPE_DEFAULT,
                            keepalive_seconds=0):
        """Create an AAPMessage for the CONFIGURE operation."""
        self.agent_id = agent_id or self._generate_agent_id()
        eid = (
            f"{self.eid_prefix}.{self.agent_id}"
            if self.is_ipn_eid
            else f"{self.eid_prefix}{self.agent_id}"
        )
        if secret is None:
            secret = str(uuid.uuid4())
        config_msg = aap2_pb2.ConnectionConfig(
            auth_type=auth_type,
            is_subscriber=subscribe,
            endpoint_id=eid,
            secret=secret,
            keepalive_seconds=keepalive_seconds,
        )
        logger.debug(f"Sending CONFIGURE message for '{agent_id}'...")
        return aap2_pb2.AAPMessage(config=config_msg), secret

    def _check_success(self, response):
        """Check that the provided AAPResponse reports success."""
        if (response.response_status !=
                aap2_pb2.ResponseStatus.RESPONSE_STATUS_SUCCESS):
            raise AAP2OperationFailed(
                "The server returned an invalid response status: " +
                str(response.response_status)
            )
        logger.debug("Request confirmed!")


class AAP2SyncClient(AAP2Client):
    """Provides common functions of the synchronous AAP2 clients."""

    def __init__(self, address):
        super().__init__(address)
        self.socket = None

    @abc.abstractmethod
    def connect(self):
        """Establish a socket connection to a uD3TN instance and return the
        received welcome message.

        Raises:
            AAP2CommunicationError: On issues receiving data.
            AAP2UnexpectedMessage: If the wrong data was received.
            protobuf.message.DecodeError: If parsing the message fails.
            OSError: If socket communication fails.

        """
        logger.debug("Connected to uD3TN, awaiting WELCOME message...")
        return self._welcome()

    def disconnect(self):
        """Shutdown and close the socket."""
        logger.debug("Terminating connection...")
        try:
            self.socket.shutdown(socket.SHUT_RDWR)
        except OSError:
            pass
        self.socket.close()

    def __enter__(self):
        """Return `self` upon calling `self.connect()` to establish the socket
        connection. See the documentation of the `connect` method.
        """
        self.connect()
        return self

    def __exit__(self, type, value, traceback):
        """Invoke `self.disconnect()` on any raised runtime exception."""
        self.disconnect()

    def _welcome(self):
        """Receive the AAP welcome message and store the node EID of the uD3TN
        instance in `self.node_eid`.
        """
        self._check_first_byte(self.socket.recv(1))
        return self._process_welcome(self.receive_msg())

    def configure(self, agent_id=None, subscribe=False, secret=None,
                  auth_type=aap2_pb2.AuthType.AUTH_TYPE_DEFAULT,
                  keepalive_seconds=0):
        """Attempt to configure the connection as specified.

        Args:
            agent_id: The agent identifier to be registered. If None,
                uuid.uuid4() is called to generate one.
            subscribe: Whether to subscribe for bundles (be a passive client)
                or not (be an active client, for sending bundles).
            secret: The AAP 2.0 authentication secret: A value that has to
                match for multiple connections re-using the same agent ID.
                If not specified, a random value will be generated.
            auth_type: The requested AAP 2.0 authentication type.
            keepalive_seconds: The interval in which keepalive messages are
                expected by uD3TN for active clients or will be sent by uD3TN
                to passive clients. Note that uD3TN will close the connection
                if an active client does not send a keepalive message within
                twice this interval. Set to zero (default value) to disable.
        Return:
            The secret that can be used for registering additional connections.

        Raises:
            AAP2CommunicationError: On issues receiving data.
            AAP2ServerDisconnected: If µD3TN disconnected in the meantime.
            AAP2OperationFailed: If uD3TN indicates a processing error.
            protobuf.message.DecodeError: If parsing the response fails.
            OSError: If socket communication fails.

        """
        config_msg, secret = self._prepare_config_msg(
            agent_id,
            subscribe,
            secret,
            auth_type,
            keepalive_seconds,
        )
        self.send(config_msg)
        response = self.receive_response()
        self._check_success(response)
        return secret

    def _receive_all(self, count):
        # Receive exactly `count` bytes from socket.
        buf = bytearray(count)
        mv = memoryview(buf)
        i = 0
        while i < count:
            rv = self.socket.recv_into(mv[i:], count - i)
            if rv == 0:
                raise AAP2ServerDisconnected("Server disconnected on `recv()`")
            elif rv < 0:
                raise AAP2CommunicationError("Communication error on `recv()`")
            i += rv
        return bytes(buf)

    def _receive_delimited(self):
        """Receive and return a byte array for which the length is indicated
        in a preceding Protobuf varint.
        """
        # Read varint
        PROTOBUF_VARINT_MAX_BYTES = 10  # 64 bits will encode as 10 bytes.
        c = 0
        result = bytearray()
        while c < PROTOBUF_VARINT_MAX_BYTES:
            data = self.socket.recv(1)
            if not data:
                raise AAP2ServerDisconnected("Server disconnected on `recv()`")
            result.append(data[0])
            # No continuation bit set -> exit loop
            if (data[0] & 0x80) == 0:
                break
            c += 1
        if c >= PROTOBUF_VARINT_MAX_BYTES:
            raise AAP2CommunicationError("Invalid varint received")
        data_len = _DecodeVarint32(result, 0)[0]
        # Read and return data
        return self._receive_all(data_len)

    def receive_msg(self):
        """Receive and return the next `AAPMessage`.

        Return:
            The AAPMessage received from uD3TN.

        Raises:
            AAP2CommunicationError: On issues receiving data.
            AAP2ServerDisconnected: If µD3TN disconnected in the meantime.
            protobuf.message.DecodeError: If parsing the message fails.
            OSError: If socket communication fails.

        """
        msg_bin = self._receive_delimited()
        msg = aap2_pb2.AAPMessage()
        msg.ParseFromString(msg_bin)
        return msg

    def receive_response(self):
        """Receive and return the next `AAPResponse`.

        Return:
            The AAPResponse received from uD3TN.

        Raises:
            AAP2CommunicationError: On issues receiving data.
            AAP2ServerDisconnected: If µD3TN disconnected in the meantime.
            protobuf.message.DecodeError: If parsing the message fails.
            OSError: If socket communication fails.

        """
        msg_bin = self._receive_delimited()
        resp = aap2_pb2.AAPResponse()
        resp.ParseFromString(msg_bin)
        return resp

    def receive_adu(self, adu_msg=None):
        """Receive a bundle ADU (Protobuf with extra payload data).

        Args:
            adu_msg (aap_pb2.BundleADU): If the `BundleADU` message was already
                received, it must be passed using this argument. Otherwise,
                the function will call `receive_msg` to receive the next
                message, which must be of type `BundleADU`.

        Return:
            A tuple of the form (BundleADU, data) containing the received
            `BundleADU` message as metadata describing the ADU payload, and
            the ADU binary payload itself.

        Raises:
            AAP2UnexpectedMessage: If the message received was not of type
                `BundleADU` (if `adu_msg` was set to `None`).
            AAP2CommunicationError: On issues receiving data.
            AAP2ServerDisconnected: If µD3TN disconnected in the meantime.
            protobuf.message.DecodeError: If parsing the message fails.
            OSError: If socket communication fails.

        """
        if not adu_msg:
            msg = self.receive_msg()
            if msg.WhichOneof("msg") != "adu":
                raise AAP2UnexpectedMessage(
                    "Expected the 'adu' oneof field but received: " +
                    msg.WhichOneof("msg"),
                    msg
                )
            adu_msg = msg.adu
        bundle_len = adu_msg.payload_length
        bundle_data = self.socket.recv(bundle_len)
        return adu_msg, bundle_data

    def send(self, msg):
        """Serialize and send the provided Protobuf message including header.

        Args:
            msg: The Protobuf message to be sent. Can be, e.g., an `AAPMessage`
                or an `AAPResponse`.

        Raises:
            protobuf.message.EncodeError: If `msg` isn't initialized.
            OSError: If socket communication fails.

        """
        msg_bin = msg.SerializeToString()
        delimited_msg = _VarintBytes(len(msg_bin)) + msg_bin
        self.socket.send(delimited_msg)

    def send_adu(self, adu_msg, bundle_data):
        """Send a bundle ADU (Protobuf with extra payload data).

        Args:
            adu_msg (aap2_pb2.BundleADU): The metadata describing the ADU in
                the form of a `BundleADU` message.
            bundle_data (bytes): The binary payload data to be sent.

        Raises:
            ValueError: If the length of `bundle_data` does not match the
                `payload_data` field of `adu_msg`.
            protobuf.message.EncodeError: If `adu_msg` isn't initialized.
            OSError: If socket communication fails.

        """
        if adu_msg.payload_length != len(bundle_data):
            raise ValueError(
                "Payload length in message does not match length of data"
            )
        self.send(aap2_pb2.AAPMessage(adu=adu_msg))
        self.socket.send(bundle_data)

    def send_response_status(self, status):
        """Send an AAPResponse with the specified status code.

        Args:
            status (aap2_pb2.ResponseStatus): The status code to be sent.

        Raises:
            OSError: If socket communication fails.

        """
        response = aap2_pb2.AAPResponse(response_status=status)
        self.send(response)


class AAP2UnixClient(AAP2SyncClient):
    """A context manager class for connecting to uD3TN's AAP 2.0 Unix socket.

    Args:
        address: The address (PATH) of the remote socket to be used when
            calling `socket.connect()`
    """

    def __init__(self, address='ud3tn.socket'):
        super().__init__(address=address)

    def connect(self):
        self.socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.socket.connect(self.address)
        return super().connect()


class AAP2TCPClient(AAP2SyncClient):
    """A context manager class for connecting to uD3TN's AAP 2.0 TCP socket.

    Args:
        address: The address tuple (HOST, PORT) of the remote socket to be used
            when calling `socket.connect()`
    """

    def __init__(self, address=('localhost', '4242')):
        super().__init__(address=address)

    def connect(self):
        host, service = self.address
        addrinfo = socket.getaddrinfo(host, service,
                                      socket.AF_UNSPEC, socket.SOCK_STREAM)
        for af, socktype, proto, canonname, sa in addrinfo:
            try:
                s = socket.socket(af, socktype, proto)
            except OSError:
                s = None
                continue
            try:
                s.connect(sa)
            except OSError:
                s.close()
                s = None
                continue
            break
        if s is None:
            raise RuntimeError(f"cannot connect to {host}:{service}")
        self.socket = s
        return super().connect()


class AAP2AsyncClient(AAP2Client):
    """Provides common functions of the asynchronous AAP2 clients."""

    def __init__(self, address):
        super().__init__(address)

    @abc.abstractmethod
    async def connect(self):
        """Establish a socket connection to a uD3TN instance and return the
        received welcome message.

        Raises:
            AAP2CommunicationError: On issues receiving data.
            AAP2UnexpectedMessage: If the wrong data was received.
            protobuf.message.DecodeError: If parsing the message fails.
            asyncio.IncompleteReadError: If the socket is closed unexpectedly.

        """
        logger.debug("Connected to uD3TN, awaiting WELCOME message...")
        self.lock = asyncio.Lock()
        return await self._welcome()

    async def disconnect(self):
        """Shutdown and close the socket."""
        logger.debug("Terminating connection...")
        async with self.lock:
            self.writer.close()
            await self.writer.wait_closed()

    async def __aenter__(self):
        """Return `self` upon calling `self.connect()` to establish the socket
        connection. See the documentation of the `connect` method.
        """
        await self.connect()
        return self

    async def __aexit__(self, type, value, traceback):
        """Invoke `self.disconnect()` on any raised runtime exception."""
        await self.disconnect()

    async def _welcome(self):
        """Receive the AAP welcome message and store the node EID of the uD3TN
        instance in `self.node_eid`.
        """
        async with self.lock:
            self._check_first_byte(await self.reader.readexactly(1))
            return self._process_welcome(await self.receive_msg())

    async def configure(self, agent_id=None, subscribe=False, secret=None,
                        auth_type=aap2_pb2.AuthType.AUTH_TYPE_DEFAULT,
                        keepalive_seconds=0):
        """Attempt to configure the connection as specified.

        Args:
            agent_id: The agent identifier to be registered. If None,
                uuid.uuid4() is called to generate one.
            subscribe: Whether to subscribe for bundles (be a passive client)
                or not (be an active client, for sending bundles).
            secret: The AAP 2.0 authentication secret: A value that has to
                match for multiple connections re-using the same agent ID.
                If not specified, a random value will be generated.
            auth_type: The requested AAP 2.0 authentication type.
            keepalive_seconds: The interval in which keepalive messages are
                expected by uD3TN for active clients or will be sent by uD3TN
                to passive clients. Note that uD3TN will close the connection
                if an active client does not send a keepalive message within
                twice this interval. Set to zero (default value) to disable.
        Return:
            The secret that can be used for registering additional connections.

        Raises:
            AAP2CommunicationError: On issues receiving data.
            AAP2ServerDisconnected: If µD3TN disconnected in the meantime.
            AAP2OperationFailed: If uD3TN indicates a processing error.
            protobuf.message.DecodeError: If parsing the response fails.
            asyncio.IncompleteReadError: If the socket is closed unexpectedly.

        """
        config_msg, secret = self._prepare_config_msg(
            agent_id,
            subscribe,
            secret,
            auth_type,
            keepalive_seconds,
        )
        async with self.lock:
            await self.send(config_msg)
            response = await self.receive_response()
        self._check_success(response)
        return secret

    async def _receive_delimited(self):
        """Receive and return a byte array for which the length is indicated
        in a preceding Protobuf varint.
        """
        # Read varint
        PROTOBUF_VARINT_MAX_BYTES = 10  # 64 bits will encode as 10 bytes.
        c = 0
        result = bytearray()
        while c < PROTOBUF_VARINT_MAX_BYTES:
            try:
                data = await self.reader.readexactly(1)
            except asyncio.exceptions.IncompleteReadError:
                raise AAP2ServerDisconnected("Server disconnected")
            result.append(data[0])
            # No continuation bit set -> exit loop
            if (data[0] & 0x80) == 0:
                break
            c += 1
        if c >= PROTOBUF_VARINT_MAX_BYTES:
            raise AAP2CommunicationError("Invalid varint received")
        data_len = _DecodeVarint32(result, 0)[0]
        # Read and return data
        try:
            return await self.reader.readexactly(data_len)
        except asyncio.exceptions.IncompleteReadError:
            raise AAP2ServerDisconnected("Server disconnected")

    async def receive_msg(self):
        """Receive and return the next `AAPMessage`.

        Note that this function is not concurrency-safe.

        Return:
            The AAPMessage received from uD3TN.

        Raises:
            AAP2CommunicationError: On issues receiving data.
            AAP2ServerDisconnected: If µD3TN disconnected in the meantime.
            protobuf.message.DecodeError: If parsing the message fails.
            asyncio.exceptions.CancelledError: If receiving is cancelled (e.g.,
                on program termination).

        """
        msg_bin = await self._receive_delimited()
        msg = aap2_pb2.AAPMessage()
        msg.ParseFromString(msg_bin)
        return msg

    async def receive_response(self):
        """Receive and return the next `AAPResponse`.

        Note that this function is not concurrency-safe.

        Return:
            The AAPResponse received from uD3TN.

        Raises:
            AAP2CommunicationError: On issues receiving data.
            AAP2ServerDisconnected: If µD3TN disconnected in the meantime.
            protobuf.message.DecodeError: If parsing the message fails.
            asyncio.exceptions.CancelledError: If receiving is cancelled (e.g.,
                on program termination).

        """
        msg_bin = await self._receive_delimited()
        resp = aap2_pb2.AAPResponse()
        resp.ParseFromString(msg_bin)
        return resp

    async def receive_adu(self, adu_msg=None):
        """Receive a bundle ADU (Protobuf with extra payload data).

        Note that this function is not concurrency-safe.

        Args:
            adu_msg (aap_pb2.BundleADU): If the `BundleADU` message was already
                received, it must be passed using this argument. Otherwise,
                the function will call `receive_msg` to receive the next
                message, which must be of type `BundleADU`.

        Return:
            A tuple of the form (BundleADU, data) containing the received
            `BundleADU` message as metadata describing the ADU payload, and
            the ADU binary payload itself.

        Raises:
            AAP2UnexpectedMessage: If the message received was not of type
                `BundleADU` (if `adu_msg` was set to `None`).
            AAP2CommunicationError: On issues receiving data.
            AAP2ServerDisconnected: If µD3TN disconnected in the meantime.
            protobuf.message.DecodeError: If parsing the message fails.
            asyncio.exceptions.CancelledError: If receiving is cancelled (e.g.,
                on program termination).

        """
        if not adu_msg:
            msg = await self.receive_msg()
            if msg.WhichOneof("msg") != "adu":
                raise AAP2UnexpectedMessage(
                    "Expected the 'adu' oneof field but received: " +
                    msg.WhichOneof("msg"),
                    msg
                )
            adu_msg = msg.adu
        bundle_len = adu_msg.payload_length
        bundle_data = await self.reader.readexactly(bundle_len)
        return adu_msg, bundle_data

    def _write_msg(self, msg):
        msg_bin = msg.SerializeToString()
        delimited_msg = _VarintBytes(len(msg_bin)) + msg_bin
        self.writer.write(delimited_msg)

    async def send(self, msg):
        """Serialize and send the provided Protobuf message including header.

        Note that this function is not concurrency-safe.

        Args:
            msg: The Protobuf message to be sent. Can be, e.g., an `AAPMessage`
                or an `AAPResponse`.

        Raises:
            protobuf.message.EncodeError: If `msg` isn't initialized.
            ConnectionResetError: If the socket is closed unexpectedly.

        """
        self._write_msg(msg)
        await self.writer.drain()

    async def send_adu(self, adu_msg, bundle_data):
        """Send a bundle ADU (Protobuf with extra payload data).

        Note that this function is not concurrency-safe.

        Args:
            adu_msg (aap2_pb2.BundleADU): The metadata describing the ADU in
                the form of a `BundleADU` message.
            bundle_data (bytes): The binary payload data to be sent.

        Raises:
            ValueError: If the length of `bundle_data` does not match the
                `payload_data` field of `adu_msg`.
            protobuf.message.EncodeError: If `adu_msg` isn't initialized.
            ConnectionResetError: If the socket is closed unexpectedly.

        """
        if adu_msg.payload_length != len(bundle_data):
            raise ValueError(
                "Payload length in message does not match length of data"
            )
        self._write_msg(aap2_pb2.AAPMessage(adu=adu_msg))
        self.writer.write(bundle_data)
        await self.writer.drain()

    async def send_rpc(self, msg):
        """Equivalent to calling send() and receive_response() with a lock.

        Args:
            msg: The Protobuf message to be sent. Can be, e.g., an `AAPMessage`
                or an `AAPResponse`.

        Return:
            The AAPResponse received from uD3TN.

        Raises:
            protobuf.message.EncodeError: If `msg` isn't initialized.
            ConnectionResetError: If the socket is closed unexpectedly.
            AAP2CommunicationError: On issues receiving data.
            AAP2ServerDisconnected: If µD3TN disconnected in the meantime.
            protobuf.message.DecodeError: If parsing the message fails.
            asyncio.exceptions.CancelledError: If receiving is cancelled (e.g.,
                on program termination).

        """
        async with self.lock:
            self._write_msg(msg)
            await self.writer.drain()
            return await self.receive_response()

    async def send_adu_rpc(self, adu_msg, bundle_data):
        """Equivalent to calling send_adu() and receive_response() with a lock.

        Note that this function is not concurrency-safe.

        Args:
            adu_msg (aap2_pb2.BundleADU): The metadata describing the ADU in
                the form of a `BundleADU` message.
            bundle_data (bytes): The binary payload data to be sent.

        Return:
            The AAPResponse received from uD3TN.

        Raises:
            ValueError: If the length of `bundle_data` does not match the
                `payload_data` field of `adu_msg`.
            protobuf.message.EncodeError: If `adu_msg` isn't initialized.
            ConnectionResetError: If the socket is closed unexpectedly.
            AAP2CommunicationError: On issues receiving data.
            AAP2ServerDisconnected: If µD3TN disconnected in the meantime.
            protobuf.message.DecodeError: If parsing the message fails.
            asyncio.exceptions.CancelledError: If receiving is cancelled (e.g.,
                on program termination).

        """
        if adu_msg.payload_length != len(bundle_data):
            raise ValueError(
                "Payload length in message does not match length of data"
            )
        async with self.lock:
            self._write_msg(aap2_pb2.AAPMessage(adu=adu_msg))
            self.writer.write(bundle_data)
            await self.writer.drain()
            return await self.receive_response()

    async def send_response_status(self, status):
        """Send an AAPResponse with the specified status code.

        Note that this function is not concurrency-safe.

        Args:
            status (aap2_pb2.ResponseStatus): The status code to be sent.

        """
        response = aap2_pb2.AAPResponse(response_status=status)
        await self.send(response)


class AAP2AsyncUnixClient(AAP2AsyncClient):
    """An asynchronous client for connecting to uD3TN's AAP 2.0 Unix socket.

    Args:
        address: The address (PATH) of the remote socket.
    """

    def __init__(self, address='ud3tn.socket'):
        super().__init__(address=address)

    async def connect(self):
        self.reader, self.writer = await asyncio.open_unix_connection(
            self.address
        )
        return await super().connect()


class AAP2AsyncTCPClient(AAP2AsyncClient):
    """An asynchronous client for connecting to uD3TN's AAP 2.0 TCP socket.

    Args:
        address: The address tuple (HOST, PORT) of the remote socket.
    """

    def __init__(self, address=('localhost', '4242')):
        super().__init__(address=address)

    async def connect(self):
        host, service = self.address
        self.reader, self.writer = await asyncio.open_connection(host, service)
        return await super().connect()



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause OR Apache-2.0
# encoding: utf-8

import argparse
import logging

from ud3tn_utils.aap2 import (
    AAP2TCPClient,
    AAP2UnixClient,
    AuthType,
    BundleADU,
    BundleADUFlags,
    ResponseStatus,
)
from ud3tn_utils.config import ConfigMessage, make_contact
from ud3tn_utils.aap2.bin.helpers import (
    add_common_parser_arguments,
    get_config_eid,
    get_secret_from_args,
    initialize_logger,
)


logger = logging.getLogger(__name__)


def main():
    import sys

    parser = argparse.ArgumentParser(
        description="configure a contact via uD3TN's AAP 2.0 interface",
    )
    add_common_parser_arguments(parser)
    parser.add_argument(
        "--dest_eid",
        default=None,
        help="the EID of the node to which the configuration belongs",
    )
    parser.add_argument(
        "eid",
        help="the EID of the node to which the contact exists",
    )
    parser.add_argument(
        "cla_address",
        help="the CLA address of the node",
    )
    parser.add_argument(
        "-s", "--schedule",
        nargs=3,
        type=int,
        metavar=("start_offset", "duration", "bitrate"),
        action="append",
        default=[],
        help="schedule a contact relative to the current time",
    )
    parser.add_argument(
        "-r", "--reaches",
        type=str,
        action="append",
        default=[],
        help="specify an EID reachable via the node",
    )
    args = parser.parse_args()
    global logger
    logger = initialize_logger(args.verbosity)

    if not args.schedule:
        logger.fatal("At least one -s/--schedule argument must be given!")
        sys.exit(1)

    msg = bytes(ConfigMessage(
        args.eid,
        args.cla_address,
        contacts=[
            make_contact(*contact)
            for contact in args.schedule
        ],
        reachable_eids=args.reaches,
    ))

    logger.debug("> %s", msg)

    if args.tcp:
        aap2_client = AAP2TCPClient(address=args.tcp)
    else:
        aap2_client = AAP2UnixClient(address=args.socket)
    with aap2_client:
        secret = aap2_client.configure(
            args.agentid,
            subscribe=False,
            secret=get_secret_from_args(args),
            auth_type=AuthType.AUTH_TYPE_BUNDLE_DISPATCH,
        )
        logger.info("Assigned agent secret: '%s'", secret)
        dest_eid = args.dest_eid or aap2_client.node_eid
        aap2_client.send_adu(
            BundleADU(
                dst_eid=get_config_eid(dest_eid),
                payload_length=len(msg),
                adu_flags=[BundleADUFlags.BUNDLE_ADU_WITH_BDM_AUTH],
            ),
            msg,
        )
        assert (
            aap2_client.receive_response().response_status ==
            ResponseStatus.RESPONSE_STATUS_SUCCESS
        )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause OR Apache-2.0
# encoding: utf-8

"""
A tool to send "ping bundles" (like the ICMP echos we all know and love)
to another DTN node, expect responses from there, and measure the time in
between.
"""

import argparse
import logging
import os
import signal
import sys
import time
import threading

from ud3tn_utils.aap import AAPUnixClient, AAPTCPClient, AAPMessage
from ud3tn_utils.aap.aap_message import AAPMessageType
from ud3tn_utils.aap.bin.helpers import (
    add_common_parser_arguments,
    initialize_logger,
)


logger = logging.getLogger(__name__)


class PingTerminated(Exception):
    """An Exception used to signal termination of the ping."""

    def __init__(self, exit_status=0):
        self.exit_status = exit_status


def _send_pings(aap_client, destination, interval, count, stop_event, timeout):

    counter = 0
    while counter < count or not count:
        # Send a bundle containing the current count & timestamp as a string.
        # Why the timestamp? -> To easily calculate how long it took to come
        # back! Otherwise, we would need some tricks for making the other
        # thread know when we sent which bundle.
        # We do not use aap_client.send but the lower-level function as
        # AAPClient invokes receive() at the end and this would confuse
        # the other thread if run here (we would receive partial data
        # everywhere).
        aap_client.socket.send(
            AAPMessage(
                AAPMessageType.SENDBUNDLE,
                destination,
                f"PING:{str(counter)}:{str(time.time())}".encode("utf-8"),
            ).serialize()
        )

        # We could use time.sleep() here, but this would not allow to properly
        # stop this thread. Thus, we rather wait until the stop Event occurs.
        # -> In normal operation this will not be the case and the function
        #    will run into the timeout in which case the condition is False.
        if stop_event.wait(timeout=interval):
            # We were requested to stop -> terminate by returning here
            return

        counter += 1

    # If we come here, termination has been requested after sending a
    # specified number of bundles and this was reached. We wait at maximum
    # for the interval to pass _again_ but do not send an additional bundle
    # during that period. Note that the receiver might trigger the stop event
    # if the last bundle has been received already.
    if timeout and stop_event.wait(timeout=timeout):
        # We were requested to stop -> no need to signal the main thread
        return

    # Terminate the main thread via KeyboardInterrupt
    os.kill(os.getpid(), signal.SIGINT)


def _try_receive_ping(aap_client, logger):
    # Wait for the next AAP message to be received
    msg = aap_client.receive()
    # Store the time we received the bundle as early as possible
    recv_time = time.time()

    # AAPClient.receive will return None, e.g., if uD3TN disconnects.
    if not msg:
        logger.warning("µD3TN has closed the connection.")
        raise PingTerminated(1)

    if msg.msg_type != AAPMessageType.RECVBUNDLE:
        # Nothing we want, continue loop.
        return False, None

    if msg.payload[0:4] != b"PING":
        # Just show we got something we do not want
        logger.warning((
            "Received bundle of length %d bytes from %s that does not seem "
            "to be a PING bundle!"
        ), len(msg.payload), msg.eid)
        return False, None

    # Try to decode the payload of the message
    try:
        _, counter_str, time_str = msg.payload.decode("utf-8").split(":")
    except UnicodeError:  # if str.decode fails
        logger.warning("Could not decode PING bundle from %s", msg.eid)
        return False, None

    # Try to get the numbers back
    try:
        counter_at_src = int(counter_str)
        time_at_src = float(time_str)
    except ValueError:  # int(...) or  float(...) failed
        logger.warning("Could not read values in PING bundle from %s", msg.eid)
        return False, None

    # Calculate the Round Trip Time (duration from sending to receiving)
    rtt = recv_time - time_at_src
    logger.info("Received PING from %s: seq=%d, rtt=%f",
                msg.eid, counter_at_src, rtt)

    return True, counter_at_src


def run_aap_ping(aap_client, destination, interval, count, logger, timeout):

    start_time = time.time()  # remember to calculate how many bdl. we expected

    # Start a second thread for sending the bundles!
    # Note that normally it is not a great idea to use threads in Python
    # due to the "Global Interpreter Lock" but in this case the threads mostly
    # wait for I/O (send/recv on the socket) which happens outside of the GIL.
    # Also note that the AAP client is not thread safe (it would call receive
    # on the socket twice concurrently and, thus, receive only partial garbage
    # in both threads). Hence, we directly use the socket send function
    # to make it thread safe...
    stop_event = threading.Event()  # for clean termination
    send_worker = threading.Thread(
        target=_send_pings,
        args=(aap_client, destination, interval, count, stop_event, timeout),
    )
    send_worker.start()

    receive_counter = 0
    try:
        while True:
            success, seqno = _try_receive_ping(aap_client, logger)
            if success:
                receive_counter += 1
                # If we got the last bundle, terminate immediately
                if count and seqno == count - 1:
                    raise PingTerminated
    # Exception handler: Either Ctrl+C was pressed or the sender signaled us a
    # SIGINT (raising a KeyboardInterrupt), or we are terminating normally.
    except (KeyboardInterrupt, PingTerminated) as e:
        # Calculate and print some statistics
        duration = time.time() - start_time
        expected_bundles = int(duration / interval) + 1
        if count:
            expected_bundles = min(expected_bundles, count)
        logger.info("Ping ran for %f seconds, received %d of %d sent",
                    duration, receive_counter, expected_bundles)
        if receive_counter < expected_bundles:
            # Indicate that we did not receive everything
            sys.exit(1)
        if isinstance(e, PingTerminated) and e.exit_status != 0:
            # Another exit status was requested
            sys.exit(e.exit_status)
    finally:  # Note that _try_receive_ping might raise SystemExit
        # Tell the sending thread to terminate
        stop_event.set()


def main():
    parser = argparse.ArgumentParser(
        description=(
            "send bundles to a given EID, wait for responses, "
            "and print them"
        ),
    )

    add_common_parser_arguments(parser)

    parser.add_argument(
        "destination",
        help="the destination EID to ping",
    )
    parser.add_argument(
        "-i", "--interval",
        type=float,
        default=1.0,
        help=(
            "interval, in seconds, to wait between sending bundles "
            "(default: 1 second)"
        ),
    )
    parser.add_argument(
        "-c", "--count",
        type=int,
        default=0,
        help=(
            "stop after sending the specified number of bundles "
            "(default: 0 = infinite)"
        ),
    )
    parser.add_argument(
        "-t", "--timeout",
        type=int,
        default=0,
        help=(
            "add a waiting time after sending all ping bundles "
            "(default: 0 = program terminates after sending all bundles)"
        ),
    )

    args = parser.parse_args()
    global logger
    logger = initialize_logger(args.verbosity + 1)  # log INFO by default

    if args.tcp:
        addr = (args.tcp[0], int(args.tcp[1]))
        with AAPTCPClient(address=addr) as aap_client:
            aap_client.register(args.agentid)
            run_aap_ping(aap_client, args.destination, args.interval,
                         args.count, logger, args.timeout)
    else:
        with AAPUnixClient(address=args.socket) as aap_client:
            aap_client.register(args.agentid)
            run_aap_ping(aap_client, args.destination, args.interval,
                         args.count, logger, args.timeout)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause OR Apache-2.0
# encoding: utf-8

"""
A tool that reads a contact plan in ION format
and configures the uD3TN nodes.
"""

import argparse
from datetime import datetime
import sys

from ud3tn_utils.aap import AAPTCPClient, AAPUnixClient
from ud3tn_utils.config import ConfigMessage, unix2dtn, Contact
from ud3tn_utils.aap.bin.helpers import get_config_eid


def get_nodes_data(nodesData):
    """Reads text file with nodes' addresses and adds them to dictionary

    Args:
        nodesData: Path to nodes configuration .txt-file
    Returns:
        nodes (dict): dictionary with each nodes and its addresses
    """
    textReader = open(nodesData, "r")
    nodesData = textReader.readlines()
    textReader.close()

    # dictionary in format name of the node : (cla-adress, (ip, port)/socket)
    # if aap connects via tcp then (cla-address, (ip, port))
    # if aap connects via socket then (cla-address, socketpath)
    nodes = {}

    for line in nodesData:
        # skipping empty lines
        if len(line) > 1:
            nodeInfo = line.split()
        else:
            continue
        if (int(nodeInfo[0]) < 1 or int(nodeInfo[0]) > 2 ** 64 - 1):
            raise Exception(
                "The given node number doesn't match the " +
                "required format. Should be: unsigned 64bit number; " +
                f"Was: {nodeInfo[0]}")

        if nodeInfo[2] == "tcp":
            nodes[nodeInfo[0]] = (nodeInfo[1], (nodeInfo[3], int(nodeInfo[4])))
        else:
            nodes[nodeInfo[0]] = (nodeInfo[1], nodeInfo[3])

    return nodes


def create_contact(start, end, bitrate):
    """Creates a `Contact` tuple for the ConfigMessage converts
    absolute or relative times into DTN timestamps

    Args:
        start (str): Start point of the contact
        end (str): End point of the contact
        bitrate (str): Bitrate of the contact
    Returns:
        Contact: tuple with DTN timestamps
    """
    if start.startswith("+"):
        start_time = unix2dtn(datetime.timestamp(datetime.now()) + int(start))
        end_time = unix2dtn(datetime.timestamp(datetime.now()) + int(end))
        return Contact(
            start=int(round(start_time)),
            end=int(round(end_time)),
            bitrate=int(bitrate),
        )
    else:
        start_time = datetime.strptime(start, "%Y/%m/%d-%H:%M:%S")
        end_time = datetime.strptime(end, "%Y/%m/%d-%H:%M:%S")

        if start_time < datetime.now() or end_time < datetime.now():
            print("Warning! The planned time for the contact has passed.")

        return Contact(
            start=int(round(unix2dtn(datetime.timestamp(start_time)))),
            end=int(round(unix2dtn(datetime.timestamp(end_time)))),
            bitrate=int(bitrate),
        )


# Converts node's name into ipn EID scheme format
def to_ipn(nodeNumber: str):
    return "ipn:" + str(nodeNumber) + ".0"


def configure_nodes(contactPlan, nodes: dict):
    """Configure the nodes by scheduling the contacts

    Args:
        contactPlan: Path to contact plan .txt-file
        nodes (dict): dictionary of nodes and their addresses
    """
    textReader = open(contactPlan, "r")
    contactPlan = textReader.readlines()
    textReader.close()

    for i, line in enumerate(contactPlan):

        # processing only a contact command
        if line.startswith("a contact"):
            planLine = line.split()
            if len(planLine) < 7:
                print(f"Error in contact plan on line {i + 1}:",
                      "Too few fields provided!")
                sys.exit(1)

            # exception when the nodes were not written in configuration file
            if planLine[4] not in nodes or planLine[5] not in nodes:
                raise Exception(f"Configuration of the nodes: {planLine[4]}",
                                f" or {planLine[5]} was not given")

            msg = bytes(ConfigMessage(
                to_ipn(planLine[5]),
                nodes[planLine[5]][0],
                contacts=[create_contact(planLine[2],
                                         planLine[3],
                                         planLine[6])],
            ))

            print(
                f"Establishing contact from {planLine[4]} to -> {planLine[5]}",
                f"Start time: {planLine[2]} end time: {planLine[3]}",
                f"With transmit rate: {planLine[6]} (bytes/s)"
                )

            # checking the type of aap port tcp/socket
            if isinstance(nodes[planLine[4]][1], tuple):
                addr = nodes[planLine[4]][1]
                with AAPTCPClient(address=addr) as aap_client:
                    aap_client.register()
                    dest_eid = to_ipn(planLine[4]) or aap_client.node_eid
                    aap_client.send_bundle(get_config_eid(dest_eid), msg)
            else:
                addr = nodes[planLine[4]][1]
                with AAPUnixClient(address=addr) as aap_client:
                    aap_client.register()
                    dest_eid = to_ipn(planLine[4]) or aap_client.node_eid
                    aap_client.send_bundle(get_config_eid(dest_eid), msg)

        else:
            # skipping empty lines and comments
            if line == '\n' or line.startswith('#'):
                continue
            else:
                print(f"Skipping unsupported command in line {i + 1}")
                continue


def main():
    parser = argparse.ArgumentParser(
        description="reads the text file",
    )

    parser.add_argument(
        "--plan",
        default=None,
        help="the path to the contact plan as txt file",
    )

    parser.add_argument(
        "--nodes",
        default=None,
        help="the path to the txt file with information about nodes"
    )

    print(
        "WARNING: This tool does not support new security mechanisms "
        "provided by µD3TN and, thus, has been deprecated. Please use "
        "aap2-config to configure contacts with the compatibility BDM "
        "and provide the administrative secret to both µD3TN and the "
        "configuration tool in release builds. "
        "For using this tool with the current version of µD3TN, "
        "please specify the --allow-remote-config commandline flag."
    )

    args = parser.parse_args()
    configure_nodes(args.plan, get_nodes_data(args.nodes))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause OR Apache-2.0
# encoding: utf-8

import argparse
import logging

import cbor2

from ud3tn_utils.aap import AAPTCPClient, AAPUnixClient
from ud3tn_utils.aap.bin.helpers import (
    add_common_parser_arguments,
    initialize_logger,
)
from pyd3tn.bundle7 import (BibeProtocolDataUnit, Bundle,
                            PayloadBlock, PrimaryBlock)


logger = logging.getLogger(__name__)


def build_bpdu(inner_source, inner_dest, payload):
    """Encapsulates a regular bundle with the chosen payload
    in a BIBE Administrative Record, thus forming a BIBE Bundle.

    Args:
        inner_source (str): Source EID of the encapsulated bundle
        inner_dest (str): Destination EID of the encapsulated bundle
        payload (bytes): The payload of the encapsulated bundle

    Returns:
        BPDU (bytes): The bytes making up the BPDU
    """
    inner_bundle = Bundle(
        PrimaryBlock(
            destination=inner_dest,
            source=inner_source,
        ),
        PayloadBlock(payload)
    )
    bibe_ar = BibeProtocolDataUnit(
        bundle=inner_bundle,
        transmission_id=0,
        retransmission_time=0,
        compatibility=False,
    )

    return cbor2.dumps(bibe_ar.record_data)


def main():
    import sys

    parser = argparse.ArgumentParser(
        description="send a bundle via uD3TN's AAP interface",
    )
    add_common_parser_arguments(parser)
    parser.add_argument(
        "dest_eid",
        help="the destination EID of the created bundle",
    )
    parser.add_argument(
        "PAYLOAD",
        default=None,
        nargs="?",
        help="the payload of the created bundle, (default: read from STDIN)",
    )
    parser.add_argument(
        "--bibe-source",
        default=None,
        help=("if set, the payload will be encapsulated twice using BIBE, "
              "with the source of the inner bundle set to the given EID"))
    parser.add_argument(
        "--bibe-destination",
        default=None,
        help=("if set, the payload will be encapsulated twice using BIBE, "
              "with the inner bundle addressed to the given EID"))
    args = parser.parse_args()

    if args.PAYLOAD:
        payload = args.PAYLOAD.encode("utf-8")
    else:
        payload = sys.stdin.buffer.read()
        sys.stdin.buffer.close()

    global logger
    logger = initialize_logger(args.verbosity)

    is_bibe = (
        args.bibe_source is not None or
        args.bibe_destination is not None
    )
    if is_bibe and (args.bibe_source is None or args.bibe_destination is None):
        logger.fatal("--bibe-source and --bibe-destination must both be set")
        sys.exit(1)

    if args.tcp:
        aap_client = AAPTCPClient(address=args.tcp)
    else:
        aap_client = AAPUnixClient(address=args.socket)
    with aap_client:
        aap_client.register(args.agentid)
        if not is_bibe:
            aap_client.send_bundle(args.dest_eid, payload, False)
        else:
            bpdu = build_bpdu(
                args.bibe_source,
                args.bibe_destination,
                payload,
            )
            aap_client.send_bundle(
                args.dest_eid,
                bpdu,
                is_bibe,
            )


if __name__ == "__main__":
    main()
